#include "ruby.h"
#include "narray.h"
#include "f2c.h"
#include "clapack.h"

static
VALUE rb_dsyevr(int argc, VALUE *argv, VALUE self)
{

  VALUE jobz;
  VALUE range;
  VALUE uplo;
  VALUE a;
  VALUE vl, vu;
  VALUE il, iu;
  VALUE abstol;
  VALUE w;
  VALUE z__ = Qnil;

  VALUE opts;

  char *cjobz = "N";
  char *crange = "A";
  char *cuplo = "U";
  integer cn;
  doublereal *ca, *ca_tmp;
  doublereal cvl, cvu;
  integer cil, ciu;
  doublereal cabstol;
  integer cm;
  doublereal *cw;
  doublereal *cz__ = NULL;
  integer *cisuppz;
  doublereal *cwork;
  integer clwork;
  integer *ciwork, cliwork;
  integer cinfo;

  int shape1, shape[2];
  VALUE indices[2];
  int flag = 0;

  rb_scan_args(argc, argv, "11", &a, &opts);
  if (!IsNArray(a))
    rb_raise(rb_eArgError, "a must be NArray");
  if (NA_RANK(a)!=2)
    rb_raise(rb_eArgError, "rank of 'a' must be 2");
  cn = NA_SHAPE0(a);
  if (NA_SHAPE1(a)!=cn)
    rb_raise(rb_eArgError, "shape[0] and shape[1] of 'a' must be the same");
  if (NA_TYPE(a) != NA_DFLOAT)
    a = na_change_type(a, NA_DFLOAT);
  ca = NA_PTR_TYPE(a, doublereal*);
  ca_tmp = ALLOC_N(doublereal,cn*cn);

  shape1 = cn;

  if (opts != Qnil) {
    Check_Type(opts, T_HASH);
    jobz = rb_hash_aref(opts, rb_str_new2("jobz"));
    if (jobz != Qnil){
      cjobz = StringValueCStr(jobz);
      switch (cjobz[0]) {
      case 'N':
      case 'n':
	flag = 0;
        cjobz = "N";
	break;
      case 'V':
      case 'v':
	flag = 1;
        cjobz = "V";
	break;
      default:
	rb_raise(rb_eArgError, "jobz must be 'N' or 'V'");
      }
    }

    range = rb_hash_aref(opts, rb_str_new2("range"));
    if (range != Qnil){
      crange = StringValueCStr(range);
      switch (crange[0]) {
      case 'A':
      case 'a':
        crange = "A";
	break;
      case 'V':
      case 'v':
	vl = rb_hash_aref(opts, rb_str_new2("vl"));
	vu = rb_hash_aref(opts, rb_str_new2("vu"));
	if (vl==Qnil || vu==Qnil)
	  rb_raise(rb_eArgError, "both vl and vu must be set when range == 'V'");
	cvl = NUM2DBL(vl);
	cvu = NUM2DBL(vu);
        crange = "V";
	break;
      case 'I':
      case 'i':
	il = rb_hash_aref(opts, rb_str_new2("il"));
	iu = rb_hash_aref(opts, rb_str_new2("iu"));
	if (il==Qnil || iu==Qnil)
	  rb_raise(rb_eArgError, "both il and iu must be set when range == 'I'");
	cil = NUM2INT(il);
	ciu = NUM2INT(iu);
	shape1 = ciu-cil+1;
        crange = "I";
	break;
      default:
	rb_raise(rb_eArgError, "range must be 'A', 'V' or 'I'");
      }
    }

    uplo = rb_hash_aref(opts, rb_str_new2("uplo"));
    if (uplo != Qnil) {
      cuplo = StringValueCStr(uplo);
      switch (cuplo[0]) {
      case 'U':
      case 'u':
        cuplo = "U";
	break;
      case 'L':
      case 'l':
        cuplo = "L";
	break;
      default:
	rb_raise(rb_eArgError, "uplo must be 'U' or 'L'");
      }
    }

    abstol = rb_hash_aref(opts, rb_str_new2("abstol"));
    cabstol = (abstol == Qnil) ? 0.0 : NUM2DBL(abstol);
  }

  shape[0] = cn;
  w = na_make_object(NA_DFLOAT, 1, shape, cNArray);
  cw = NA_PTR_TYPE(w, doublereal*);

  if (flag) {
    shape[0] = cn;
    shape[1] = shape1;
    z__ = na_make_object(NA_DFLOAT, 2, shape, cNArray);
    cz__ = NA_PTR_TYPE(z__, doublereal*);
  }

  cisuppz = ALLOC_N(integer, 2*cn);

  clwork = -1;
  cliwork = -1;
  cwork = ALLOC(doublereal);
  ciwork = ALLOC(integer);
  dsyevr_(cjobz, crange, cuplo,
	  &cn, ca_tmp, &cn, &cvl, &cvu, &cil, &ciu, &cabstol,
	  &cm, cw, cz__, &cn, cisuppz, cwork, &clwork, ciwork, &cliwork, &cinfo);
  clwork = (integer)cwork[0];
  cliwork = ciwork[0];
  free(cwork);
  free(ciwork);
  cwork = ALLOC_N(doublereal, clwork);
  ciwork = ALLOC_N(integer, cliwork);
  MEMCPY(ca_tmp, ca, doublereal, cn*cn);
  dsyevr_(cjobz, crange, cuplo,
	  &cn, ca_tmp, &cn, &cvl, &cvu, &cil, &ciu, &cabstol,
	  &cm, cw, cz__, &cn, cisuppz, cwork, &clwork, ciwork, &cliwork, &cinfo);
  free(cwork);
  free(ciwork);
  free(ca_tmp);
  free(cisuppz);
  if (cm < cn) {
    indices[0] = rb_range_new(INT2NUM(0), INT2NUM(cm), 1);
    w = na_aref(1,indices, w);
    if (flag && cm < shape1) {
      indices[1] = indices[0];
      indices[0] = Qtrue;
      z__ = na_aref(2, indices, z__);
    }
  }

  return flag ? rb_ary_new3(2, w, z__) : w;
}


static
VALUE rb_ssyevr(int argc, VALUE *argv, VALUE self)
{

  VALUE jobz;
  VALUE range;
  VALUE uplo;
  VALUE a;
  VALUE vl, vu;
  VALUE il, iu;
  VALUE abstol;
  VALUE w;
  VALUE z__ = Qnil;

  VALUE opts;

  char *cjobz = "N";
  char *crange = "A";
  char *cuplo = "U";
  integer cn;
  real *ca, *ca_tmp;
  real cvl, cvu;
  integer cil, ciu;
  real cabstol;
  integer cm;
  real *cw;
  real *cz__ = NULL;
  integer *cisuppz;
  real *cwork;
  integer clwork;
  integer *ciwork, cliwork;
  integer cinfo;

  int shape1, shape[2];
  VALUE indices[2];
  int flag = 0;

  rb_scan_args(argc, argv, "11", &a, &opts);
  if (!IsNArray(a))
    rb_raise(rb_eArgError, "a must be NArray");
  if (NA_RANK(a)!=2)
    rb_raise(rb_eArgError, "rank of 'a' must be 2");
  cn = NA_SHAPE0(a);
  if (NA_SHAPE1(a)!=cn)
    rb_raise(rb_eArgError, "shape[0] and shape[1] of 'a' must be the same");
  if (NA_TYPE(a) != NA_SFLOAT)
    a = na_change_type(a, NA_SFLOAT);
  ca = NA_PTR_TYPE(a, real*);
  ca_tmp = ALLOC_N(real,cn*cn);

  shape1 = cn;

  if (opts != Qnil) {
    Check_Type(opts, T_HASH);
    jobz = rb_hash_aref(opts, rb_str_new2("jobz"));
    if (jobz != Qnil){
      cjobz = StringValueCStr(jobz);
      switch (cjobz[0]) {
      case 'N':
      case 'n':
	flag = 0;
        cjobz = "N";
	break;
      case 'V':
      case 'v':
	flag = 1;
        cjobz = "V";
	break;
      default:
	rb_raise(rb_eArgError, "jobz must be 'N' or 'V'");
      }
    }

    range = rb_hash_aref(opts, rb_str_new2("range"));
    if (range != Qnil){
      crange = StringValueCStr(range);
      switch (crange[0]) {
      case 'A':
      case 'a':
        crange = "A";
	break;
      case 'V':
      case 'v':
	vl = rb_hash_aref(opts, rb_str_new2("vl"));
	vu = rb_hash_aref(opts, rb_str_new2("vu"));
	if (vl==Qnil || vu==Qnil)
	  rb_raise(rb_eArgError, "both vl and vu must be set when range == 'V'");
	cvl = (real)NUM2DBL(vl);
	cvu = (real)NUM2DBL(vu);
        crange = "V";
	break;
      case 'I':
      case 'i':
	il = rb_hash_aref(opts, rb_str_new2("il"));
	iu = rb_hash_aref(opts, rb_str_new2("iu"));
	if (il==Qnil || iu==Qnil)
	  rb_raise(rb_eArgError, "both il and iu must be set when range == 'I'");
	cil = NUM2INT(il);
	ciu = NUM2INT(iu);
	shape1 = ciu-cil+1;
        crange = "I";
	break;
      default:
	rb_raise(rb_eArgError, "range must be 'A', 'V' or 'I'");
      }
    }

    uplo = rb_hash_aref(opts, rb_str_new2("uplo"));
    if (uplo != Qnil) {
      cuplo = StringValueCStr(uplo);
      switch (cuplo[0]) {
      case 'U':
      case 'u':
        cuplo = "U";
	break;
      case 'L':
      case 'l':
        cuplo = "L";
	break;
      default:
	rb_raise(rb_eArgError, "uplo must be 'U' or 'L'");
      }
    }

    abstol = rb_hash_aref(opts, rb_str_new2("abstol"));
    cabstol = (abstol == Qnil) ? 0.0 : (real)NUM2DBL(abstol);
  }

  shape[0] = cn;
  w = na_make_object(NA_SFLOAT, 1, shape, cNArray);
  cw = NA_PTR_TYPE(w, real*);

  if (flag) {
    shape[0] = cn;
    shape[1] = shape1;
    z__ = na_make_object(NA_SFLOAT, 2, shape, cNArray);
    cz__ = NA_PTR_TYPE(z__, real*);
  }

  cisuppz = ALLOC_N(integer, 2*cn);

  clwork = -1;
  cliwork = -1;
  cwork = ALLOC(real);
  ciwork = ALLOC(integer);
  ssyevr_(cjobz, crange, cuplo,
	  &cn, ca_tmp, &cn, &cvl, &cvu, &cil, &ciu, &cabstol,
	  &cm, cw, cz__, &cn, cisuppz, cwork, &clwork, ciwork, &cliwork, &cinfo);
  clwork = (integer)cwork[0];
  cliwork = ciwork[0];
  free(cwork);
  free(ciwork);
  cwork = ALLOC_N(real, clwork);
  ciwork = ALLOC_N(integer, cliwork);
  MEMCPY(ca_tmp, ca, real, cn*cn);
  ssyevr_(cjobz, crange, cuplo,
	  &cn, ca_tmp, &cn, &cvl, &cvu, &cil, &ciu, &cabstol,
	  &cm, cw, cz__, &cn, cisuppz, cwork, &clwork, ciwork, &cliwork, &cinfo);
  free(cwork);
  free(ciwork);
  free(ca_tmp);
  free(cisuppz);
  if (cm < cn) {
    indices[0] = rb_range_new(INT2NUM(0), INT2NUM(cm), 1);
    w = na_aref(1,indices, w);
    if (flag && cm < shape1) {
      indices[1] = indices[0];
      indices[0] = Qtrue;
      z__ = na_aref(2, indices, z__);
    }
  }

  return flag ? rb_ary_new3(2, w, z__) : w;
}



void Init_lapack()
{

  VALUE mNumRu;
  VALUE mLapack;

  rb_require("narray");

  mNumRu = rb_define_module("NumRu");
  mLapack = rb_define_module_under(mNumRu, "Lapack");

  rb_define_module_function(mLapack, "dsyevr", rb_dsyevr, -1);
  rb_define_module_function(mLapack, "ssyevr", rb_ssyevr, -1);
}
